unit Moments;

interface uses SVD, FFT;

const
    mfilt = 256;

procedure ComputeMoments(var fmag: fourier_mag; nspec: Integer;
    sample_ms: Double; var mean, stdv, skew, kurt: Double);

procedure MomentModel(VAR mean, stdv, skew, kurt: Double;
    var fmag: fourier_mag; nspec: Integer; sample_ms: Double);

procedure MomentFilt(var sig: fourier_record; m: Integer;
    var filt: fourier_mag; nspec: Integer);


implementation

procedure ComputeMoments(var fmag: fourier_mag; nspec: Integer;
    sample_ms: Double; var mean, stdv, skew, kurt: Double);

    var i, j: Integer;
        fn, f, fr, mag: Double;
        m: array [0 .. 4] of Double;
begin
    fn := 1.0;
    if (nspec > 0) and (sample_ms > 0) then
        fn := 0.5/(sample_ms*nspec);
    for j := 0 to 4 do m[j] := 0;
    for i := 1 to Pred(nspec) do
        m[0] := m[0] + fmag[i];
    mag := m[0];
    m[0] := 1.0;
    for i := 1 to Pred(nspec) do begin
        f := i*fn;
        m[1] := m[1] + f*fmag[i]/mag;
    end;
    mean := m[1];
    for i := 1 to Pred(nspec) do begin
        f := i*fn - mean;
        m[2] := m[2] + f*f*fmag[i]/mag;
    end;
    stdv := Sqrt(m[2]);
    for i := 1 to Pred(nspec) do begin
        f := (i*fn - mean)/stdv;
        fr := f*f;
        for j := 3 to 4 do begin
            fr := fr*f;
            m[j] := m[j] + fr*fmag[i]/mag;
        end;
    end;
    skew := m[3];
    kurt := m[4] - 3.0;
end;

procedure MomentModel(VAR mean, stdv, skew, kurt: Double;
    var fmag: fourier_mag; nspec: Integer; sample_ms: Double);

    var i, j, k: Integer;
        fn, f, fr, p, fac, distance, distance0: Double;
        m: array [0 .. 4] of Double;
        a, a0, da: array [0 .. 4] of Double;
        u: array [0 .. 8] of Double;
        at: a_matrix;
        vt: v_matrix;
        z: z_array;
begin
    fn := 1.0;
    if (nspec > 0) and (sample_ms > 0) then
        fn := 0.5/(sample_ms*nspec);

    m[0] := 1.0;
    m[1] := 0.0;
    m[2] := 1.0;
    m[3] := skew;
    m[4] := kurt + 3.0;

    for j := 0 to 4 do a[j] := 0;
    a[0] := -Ln(nspec+1);
    for j := 0 to 4 do a0[j] := a[j];
    distance0 := 1;
    fac := 1.0;
    k := 0;
    while (k < 100) and (distance0 > 1e-6) do begin
        Inc(k);
        for j := 0 to 8 do u[j] := 0;
        for i := 0 to nspec do begin
            f := (i*fn - mean)/stdv;
            fr := 1.0;
            p := a[0];
            for j := 1 to 4 do begin
                fr := fr*f;
                p := p + a[j]*fr;
            end;
            if p < 0.0 then
                p := Exp(p)
            else
                p := 1.0;
            fmag[i] := p;
            fr := 1.0;
            u[0] := u[0] + p;
            for j := 1 to 8 do begin
                fr := fr*f;
                u[j] := u[j] + fr*p;
            end;
        end;

        distance := 0;
        for i := 0 to 4 do
            distance := distance + (m[i] - u[i])*(m[i] - u[i]);

        if (distance > distance0) and (k > 1) then begin
            fac := 0.5*fac;
            for i := 0 to 4 do a[i] := a0[i];
        end else begin
            fac := 1.4*fac;
            if fac > 1 then fac := 1;
            for i := 0 to 4 do a0[i] := a[i];
            distance0 := distance;
            for i := 1 to 5 do for j := 1 to 5 do
                at[i][j] := u[i+j-2];
            SingValDecomp(at,vt,z,5,5);
            for i := 0 to 4 do begin
                da[i] := 0.0;
                for j := 0 to 4 do
                    da[i] := da[i] + at[i+1][j+1] * (m[j] - u[j]);
                da[i] := da[i]/Sqrt(z[i+1]*z[i+1] + 1e-8*z[1]*z[1]);
            end;
        end;

        for i := 0 to 4 do begin
            for j := 0 to 4 do
                a[j] := a[j] + vt[i+1][j+1]*fac*da[i];
        end;
{
        Writeln(k:10,Sqrt(distance0):14:3,fac:10:3,
            u[0]:8:3,u[1]:8:3,u[2]:8:3,u[3]:8:3,u[4]:8:3);
 }
    end;
end;

procedure MomentFilt(var sig: fourier_record; m: Integer;
    var filt: fourier_mag; nspec: Integer);

    var i: Integer;
        gain: Double;
begin
    RealFourierTransform(sig,m);
    for i := 0 to nspec do begin
        gain := Sqrt(nspec*filt[i]);
        with sig.a[i] do begin
            re := re * gain;
            im := im * gain;
        end;
    end;
    RealInverseTransform(sig,m);
end;

end.