UNIT StrBlock; {Copyright 1994, 1998 Paul Milenkovic, 118 Shiloh Dr.,
                Madison, WI 53705-2433 U.S.A.

                Free software: you are permitted to copy, distribute, and
                utilize this software provided you include this notice.}

INTERFACE

CONST
    ASCIIZ_length = 127;

TYPE
    ASCIIZ_string = STRING[ASCIIZ_length];
    string_block  = ARRAY [0 .. 1023] OF CHAR;
    string_blockP = ^string_block;


FUNCTION  RetrieveStringBlock(
    VAR dst: STRING; VAR block: string_block; p, n: INTEGER): INTEGER;

FUNCTION  RetrieveStringBlockSuffix(
    VAR count: INTEGER; VAR block: string_block; p, n: INTEGER): INTEGER;

FUNCTION  EnvPtr: string_blockP;

FUNCTION  ComSpec: ASCIIZ_string;


PROCEDURE UpdateStringKeyword(VAR dest: ASCIIZ_string; src, key: STRING);

PROCEDURE UpdateIntegerKeyword(VAR dest: INTEGER; src, key: STRING);

PROCEDURE UpdateWordHexKeyword(VAR dest: WORD; src, key: STRING);

PROCEDURE UpdateLongintKeyword(VAR dest: LONGINT; src, key: STRING);

PROCEDURE UpdateSingleKeyWord(VAR dest: SINGLE; src, key: STRING);

PROCEDURE UpdateDoubleKeyWord(VAR dest: DOUBLE; src, key: STRING);


FUNCTION  HexString(num: WORD): ASCIIZ_string;

FUNCTION  LHexString(num: LONGINT): ASCIIZ_string;

FUNCTION  IntString(n, m: INTEGER): ASCIIZ_string;


FUNCTION  FileExists(fname: ASCIIZ_string): BOOLEAN;

FUNCTION  FileReadOnly(fname: ASCIIZ_string): BOOLEAN;

PROCEDURE EraseFile(fname: ASCIIZ_string);


FUNCTION  RootPath(path: ASCIIZ_string): ASCIIZ_string;

FUNCTION  FileInPath(path: ASCIIZ_string): ASCIIZ_string;

FUNCTION  DirInPath (path: ASCIIZ_string): ASCIIZ_string;

FUNCTION  FileInDir(path: ASCIIZ_string): ASCIIZ_string;

FUNCTION  FullPath(path: ASCIIZ_string): ASCIIZ_string;


IMPLEMENTATION

FUNCTION  RetrieveStringBlock(
    VAR dst: STRING; VAR block: string_block; p, n: INTEGER): INTEGER;

    VAR i: INTEGER;

BEGIN
    i := 0;
    WHILE (p < Pred(n)) AND (block[p] <> #00) DO BEGIN
        IF i < ASCIIZ_length THEN BEGIN
            Inc(i);
            dst[i] := block[p];
        END;
        Inc(p);
    END;
    dst[0] := CHAR(i);
    IF (i > 0) AND (p < Pred(n)) THEN
        Inc(p);
    RetrieveStringBlock := p;
END;


FUNCTION  RetrieveStringBlockSuffix(
    VAR count: INTEGER; VAR block: string_block; p, n: INTEGER): INTEGER;

    VAR i: INTEGER;

BEGIN
    count := 0;
    IF p < n - 3 THEN BEGIN
        IF block[p] = #0 THEN BEGIN
            i := BYTE(block[p+1]);
            i := i OR (BYTE(block[p+2]) SHL 8);
            IF i > 0 THEN BEGIN
                count := i;
                Inc(p,3);
            END;
        END;
    END;
    RetrieveStringBlockSuffix := p;
END;


FUNCTION  EnvPtr: string_blockP;

    VAR env_ptr: ^string_block;

BEGIN
    EnvPtr := string_blockP(Ptr(INTEGER(Ptr(PrefixSeg,$002C)^),0));
END;


FUNCTION ComSpec: ASCIIZ_string;

    VAR system_path, estr: ASCIIZ_string;
	i, p: INTEGER;

BEGIN
    system_path := '';
    p := 0;
    REPEAT
        p := RetrieveStringBlock(estr,EnvPtr^,p,MaxInt);
        UpdateStringKeyword(system_path,estr,'COMSPEC=');
    UNTIL estr = '';
    i := 1;
    WHILE i <= Length(system_path) DO BEGIN
        IF system_path[i] = ';' THEN system_path[0] := CHAR(Pred(i));
        Inc(i);
    END;
    ComSpec := system_path;
END;


PROCEDURE UpdateStringKeyword(VAR dest: ASCIIZ_string; src, key: STRING);

    VAR n: INTEGER;

BEGIN
    n := Length(key);
    IF Copy(src,1,n) = key THEN dest := Copy(src,n+1,Length(src)-n);
END;


PROCEDURE UpdateIntegerKeyword(VAR dest: INTEGER; src, key: STRING);

    VAR n, err: INTEGER;

BEGIN
    n := Length(key);
    IF Copy(src,1,n) = key THEN Val(Copy(src,n+1,Length(src)-n),dest,err);
END;


PROCEDURE UpdateWordHexKeyword(VAR dest: WORD; src, key: STRING);

    VAR i, n: INTEGER;

BEGIN
    n := Length(key);
    IF Copy(src,1,n) = key THEN BEGIN
        dest := 0;
        FOR i := Succ(n) TO Length(src) DO
            dest := (dest SHL 4) + (BYTE(src[i]) AND $0F) +
                    9*(BYTE(src[i]) SHR 6);
    END;
END;


PROCEDURE UpdateLongintKeyword(VAR dest: LONGINT; src, key: STRING);

    VAR n, err: INTEGER;

BEGIN
    n := Length(key);
    IF Copy(src,1,n) = key THEN Val(Copy(src,n+1,Length(src)-n),dest,err);
END;


PROCEDURE UpdateSingleKeyWord(VAR dest: SINGLE; src, key: STRING);

    VAR n, err: INTEGER;

BEGIN
    n := Length(key);
    IF Copy(src,1,n) = key THEN Val(Copy(src,n+1,Length(src)-n),dest,err);
END;


PROCEDURE UpdateDoubleKeyWord(VAR dest: DOUBLE; src, key: STRING);

    VAR n, err: INTEGER;

BEGIN
    n := Length(key);
    IF Copy(src,1,n) = key THEN Val(Copy(src,n+1,Length(src)-n),dest,err);
END;


FUNCTION  HexString(num: WORD): ASCIIZ_string;

    VAR sval: STRING[4];
        i: INTEGER;
        fld: WORD;

BEGIN
    sval := '';
    FOR i := 0 TO 3 DO BEGIN
        fld := num SHR 12;
        num := num SHL 4;
        IF (sval <> '') OR (fld <> 0) OR (i = 3) THEN BEGIN
            IF fld > 9 THEN
                sval := Concat(sval,CHAR(55+fld))
            ELSE
                sval := Concat(sval,CHAR(48+fld));
        END;
    END;
    HexString := sval;
END;


FUNCTION  LHexString(num: LONGINT): ASCIIZ_string;

    VAR sval: STRING[8];
        i: INTEGER;
        fld: WORD;

BEGIN
    sval := '';
    FOR i := 0 TO 7 DO BEGIN
        fld := num SHR 16;
        fld := fld SHR 12;
        num := num SHL 4;
        IF (sval <> '') OR (fld <> 0) OR (i = 7) THEN BEGIN
            IF fld > 9 THEN
                sval := Concat(sval,CHAR(55+fld))
            ELSE
                sval := Concat(sval,CHAR(48+fld));
        END;
    END;
    LHexString := sval;
END;


FUNCTION  IntString(n, m: INTEGER): ASCIIZ_string;

    VAR sval: ASCIIZ_string;

BEGIN
    Str(n:m,sval);
    IntString := sval;
END;


FUNCTION FileExists(fname: ASCIIZ_string): BOOLEAN;

    VAR tfile: FILE;

BEGIN
    FileExists := FALSE;
    Assign(tfile,fname);
    {$I-} Reset(tfile) {$I+};
    IF IOResult = 0 THEN BEGIN
        Close(tfile);
        FileExists := TRUE;
    END;
END;


FUNCTION FileReadOnly(fname: ASCIIZ_string): BOOLEAN;

    VAR tfile: FILE;

BEGIN
    Assign(tfile,fname);
    FileMode := 0;
    {$I-} Reset(tfile) {$I+};
    FileMode := 2;
    IF IOResult = 0 THEN BEGIN
        {$I-} Reset(tfile) {$I+};
        IF IOResult = 0 THEN BEGIN
            Close(tfile);
            FileReadOnly := FALSE;
        END ELSE
            FileReadOnly := TRUE;
    END ELSE BEGIN
        {$I-} Rewrite(tfile) {$I+};
        IF IOResult = 0 THEN BEGIN
            Close(tfile);
            Erase(tfile);
            FileReadOnly := FALSE;
        END ELSE
            FileReadOnly := TRUE;
    END;
END;


PROCEDURE EraseFile(fname: ASCIIZ_string);

    VAR tfile: FILE;

BEGIN
    Assign(tfile,fname);
    {$I-} Erase(tfile) {$I+};
    IF IOResult = 0 THEN;
END;


FUNCTION  RootPath(path: ASCIIZ_string): ASCIIZ_string;

    VAR p, q: INTEGER;

BEGIN
    p := Length(path);
    WHILE (p > 1) AND ((path[p] = ' ') OR (path[p] = #0)) DO Dec(p);
    q := p;
    WHILE (p > 1) AND (path[p] <> '.') AND (path[p] <> '\') DO Dec(p);
    IF path[p] = '.' THEN RootPath := Copy(path,1,Pred(p))
                     ELSE RootPath := Copy(path,1,q);
END;


FUNCTION  ParseFile(VAR fname: ASCIIZ_string): INTEGER;

    VAR p: INTEGER;

BEGIN
    ParseFile := 0;
    p := Length(fname);
    WHILE p > 0 DO
        IF (fname[p] = '\') OR (fname[p] = ':') THEN BEGIN
            ParseFile := p;
            p := 0;
        END ELSE
            Dec(p);
END;


FUNCTION  FileInPath(path: ASCIIZ_string): ASCIIZ_string;

    VAR p: INTEGER;

BEGIN
    p := ParseFile(path);
    FileInPath := Copy(path,Succ(p),Length(path)-p);
END;


FUNCTION  DirInPath (path: ASCIIZ_string): ASCIIZ_string;

    VAR p: INTEGER;

BEGIN
    p := ParseFile(path);
    IF p > 1 THEN
        IF (path[p] = '\') AND (path[Pred(p)] <> ':') THEN Dec(p);
    DirInPath := Copy(path,1,p);
END;


FUNCTION  FileInDir(path: ASCIIZ_string): ASCIIZ_string;

    VAR p: INTEGER;

BEGIN
    p := ParseFile(path);
    FileInDir := Copy(path,Succ(p),Length(path)-p);
    IF p > 1 THEN
        IF (path[p] = '\') AND (path[Pred(p)] <> ':') THEN Dec(p);
    IF p > 1 THEN
        ChDir(Copy(path,1,p));
END;


FUNCTION  FullPath(path: ASCIIZ_string): ASCIIZ_string;

    VAR new_dir, old_dir: ASCIIZ_string;

BEGIN
    GetDir(0,old_dir);
    path := FileInDir(path);
    GetDir(0,new_dir);
    IF new_dir[Length(new_dir)] = '\' THEN
        FullPath := Concat(new_dir,path)
    ELSE
        FullPath := Concat(new_dir,'\',path);
    ChDir(old_dir);
END;

END.
