PROGRAM VoiceA; {Copyright 1995, 1997 Paul Milenkovic ALL RIGHTS RESERVED}

USES StrBlock, StdIO, DSPTypes, DSPFast, LPC, FiltBuf;

CONST
    lpi_out = 1; {Output LPC inverse filter}
    cor_out = 2; {Output optimum correlator/integrator filter}

    VU_ms: DOUBLE = 8.0;

VAR
    sample_ms, start_ms, window_ms: DOUBLE;
    ipar, ncoef, out_mode: INTEGER;
    scale, alpha, data_scale: DOUBLE;
    mindb, cfx, dfx, fg, bg, bz: SINGLE;

    pstr, fname, extension: ASCIIZ_string;
    pfile: TEXT;

    wrk_buf, lpi_buf: FiltBuf.wrk_bufT;
    sig: DSPTypes.sig_buffer;


PROCEDURE ComputeCoef(VAR b0, a1, a2: SINGLE; ft, bt: SINGLE; n: INTEGER);

    VAR y0, y1, y2: SINGLE;
        i: INTEGER;

BEGIN
    a2 := Exp(-Pi*bt);
    a1 := -2*a2*Cos(2*Pi*ft);
    a2 := a2*a2;
    y2 := -a1;
    y1 := -1 + a1*a1 - a2;
    b0 := 1 + y2*y2 + y1*y1;
    FOR i := 3 TO Pred(n) DO BEGIN
        y0 := -a1*y1 - a2*y2;
        y2 := y1;
        y1 := y0;
        b0 := b0 + y0*y0;
    END;
    IF b0 > 0 THEN b0 := 1/Sqrt(b0);
END;


PROCEDURE LPCInverseFilter;

    CONST
        pmax = 16;
        kmax = 40;

    TYPE
        karray = ARRAY [1 .. kmax] OF SINGLE;

    VAR
        nread, i, k, p, ip, im: INTEGER;

        nLPC, nlp, nVU, mVU: INTEGER;
        wrk_len, wrk_top, wrk_dif: INTEGER;
        vuv_pre, vuv_pos, vuv_pst, last_uv: INTEGER;
        lpc_pos, lpi_top, itg_len, itg_top, itg_nzi, itg_zcr: INTEGER;
        tbuf: DOUBLE;

        pcoef: predictor;
        avg_db, avg_cf, max_cf, scale_db, db_scale, scale_cf: SINGLE;
        sig0, sumsq, ssqer, dmp, DC_scale, c: SINGLE;
        nzc: INTEGER;

        ppm_val: ARRAY [1 .. pmax] OF SINGLE;
        ppm_pos, ppm_km: ARRAY [1 .. pmax] OF INTEGER;
        ppm_ok: ARRAY [1 .. pmax] OF BOOLEAN;

        val_pre, val_pst: SINGLE;
        val_opt, lpisq: SINGLE;
        km_ref, pos_ref: INTEGER;
        pulse_pending, dbl_pulse: BOOLEAN;

        x0, x1, x2: SINGLE;
        y0, y1, y2, b0, a1, a2, f0: karray;

    PROCEDURE OutputPulse(p: INTEGER);
    BEGIN
        IF ppm_ok[p] THEN
            Writeln(pfile,
                tbuf+ppm_pos[p]*sample_ms:10:3,ppm_val[p]:10:6,
                f0[ppm_km[p]]:10:3,f0[ppm_km[p]]*bg:10:3,
                ppm_km[p]:3
            );
    END;

BEGIN
{Initialize buffer markers}
    nLPC := Round(window_ms/sample_ms); {Length of LPC frame}
    IF nLPC > mbuf THEN
        nLPC := mbuf;
    mVU := Round(0.5*VU_ms/sample_ms);
    nVU := Succ(2*mVU);                {Analysis length for V/UV decision}
    wrk_len := 0;                      {Samples in the buffer}
    itg_len := 0;                      {Samples ready for integration}
    lpc_pos := 0;                      {Start of LPC frame}
    lpi_top := 0;                      {Top inverse filter sample}
    itg_top := 0;                      {Top integrated sample}
    itg_nzi := 0;                      {Last non-zero integrator input}
    itg_zcr := 0;                      {Integrator zero crossing position}
    vuv_pre := 0;                      {V/UV analysis frame start}
    vuv_pos := 0;                      {V/UV analysis frame center}
    vuv_pst := 0;                      {V/UV analysis frame finish}
    last_uv := 0;                      {Last unvoiced (UV) sample}
    nzc     := 0;                      {V/UV zero crossing count}
    sumsq   := 0;                      {V/UV MS signal}
    lpisq   := 0;                      {MS glottal pulse}
    tbuf := start_ms;                  {Time of start of buffer}
    scale_db := scale*scale;
    db_scale := 10.0/Ln(10.0);
    scale_cf := 0.5/(sample_ms);

{Initialize integration (correlator) filters}
    bg := bg/fg;
    FOR k := 1 TO kmax DO BEGIN
        f0[k] := fg;
        fg := fg * 1.05;
        ComputeCoef(
            b0[k],a1[k],a2[k],
            f0[k]*sample_ms,f0[k]*bg*sample_ms,nLPC*2
        );
        y1[k] := 0;
        y2[k] := 0;
    END;
    x1 := 0;
    x2 := 0;

{Reset candidate pitch pulse markers}
    FOR p := 1 TO pmax DO BEGIN
        ppm_val[p] := 0;               {Pulse amplitude}
        ppm_pos[p] := 0;               {Pulse location}
        ppm_km [p] := 0;               {Integrating filter index 1..kmax}
        ppm_ok [p] := FALSE;           {Pulse valid/invalid}
    END;
    pulse_pending := FALSE;            {Zero crossing means possible pulse}
     km_ref := 0;                      {Most recent integrator index}
    pos_ref := 0;                      {Most recent pulse location}
    val_opt := 0.0;                    {Strongest pulse amplitude in 20 ms}

    REPEAT
{Update buffers}
        IF wrk_len > wrk_mrk THEN BEGIN
            wrk_dif := wrk_len - wrk_mrk;
            IF out_mode > 0 THEN
                StdBlockWrite(lpi_buf[1],wrk_dif*Sizeof(INTEGER));
            Move(wrk_buf[Succ(wrk_dif)],wrk_buf[1],wrk_mrk*Sizeof(INTEGER));
            Move(lpi_buf[Succ(wrk_dif)],lpi_buf[1],wrk_mrk*Sizeof(INTEGER));
            Dec(wrk_len,wrk_dif);
            Dec(lpc_pos,wrk_dif);
            Dec(lpi_top,wrk_dif);
            Dec(itg_top,wrk_dif);
            Dec(itg_nzi,wrk_dif);
            Dec(itg_zcr,wrk_dif);
            Dec(vuv_pre,wrk_dif);
            Dec(vuv_pos,wrk_dif);
            Dec(vuv_pst,wrk_dif);
            Dec(last_uv,wrk_dif);
            IF lpi_top < 0 THEN
                lpi_top := 0;
            IF itg_top < 0 THEN
                itg_top := 0;
            IF itg_nzi < 0 THEN
                itg_nzi := 0;
            IF itg_zcr < 0 THEN
                itg_zcr := 0;
            IF last_uv < 0 THEN
                last_uv := 0;
            Dec(pos_ref,wrk_dif);
            IF pos_ref < 1 THEN BEGIN
                pos_ref := 0;
                 km_ref := 0;
            END;
            tbuf := tbuf + wrk_dif*sample_ms;
            FOR p := pmax DOWNTO 1 DO BEGIN
                Dec(ppm_pos[p],wrk_dif);
                IF ppm_pos[p] < 1 THEN BEGIN
                    OutputPulse(p);
                    ppm_val[p] := 0;
                    ppm_pos[p] := 0;
                    ppm_km [p] := 0;
                    ppm_ok [p] := FALSE;
                END;
            END;
        END;
        nread := StdBlockRead(
            wrk_buf[Succ(wrk_len)],Sizeof(wrk_buf)-Sizeof(INTEGER)*wrk_len
        );
        wrk_len := wrk_len + nread DIV Sizeof(INTEGER);

{Process LPC frames within current buffer extent}
        wrk_top := wrk_len;
        IF nread > 0 THEN
            wrk_top := wrk_top - nLPC;
        WHILE vuv_pos < wrk_top DO BEGIN
            Inc(vuv_pos);
            IF lpc_pos < last_uv THEN
                lpc_pos := last_uv;
            WHILE vuv_pst < vuv_pos+mVU DO BEGIN
                Inc(vuv_pst);
                IF (vuv_pst <= wrk_len) AND (vuv_pst > 1) THEN BEGIN
                    IF (wrk_buf[vuv_pst] < 0) XOR
                       (wrk_buf[Pred(vuv_pst)] < 0)
                    THEN
                        Inc(nzc);
                    sig0 := wrk_buf[vuv_pst];
                    sumsq := sumsq + sig0*sig0;
                END;
            END;
            WHILE vuv_pre < Pred(vuv_pos-mVU) DO BEGIN
                Inc(vuv_pre);
                IF vuv_pre > 1 THEN BEGIN
                    IF (wrk_buf[vuv_pre] < 0) XOR
                       (wrk_buf[Pred(vuv_pre)] < 0)
                    THEN
                        Dec(nzc);
                    sig0 := wrk_buf[vuv_pre];
                    sumsq := sumsq - sig0*sig0;
                END;
            END;
            IF (vuv_pst > vuv_pre) AND (sumsq > 0) THEN BEGIN
                avg_cf := scale_cf*nzc/(vuv_pst-vuv_pre);
                avg_db := db_scale*Ln(scale_db*sumsq/(vuv_pst-vuv_pre));
            END ELSE BEGIN
                avg_cf := 10;
                avg_db := -100;
            END;
            IF mindb < -20.0 THEN
                cfx := 0.8 + 1.4*(avg_db - mindb)/20.0
            ELSE
                cfx := 2.2;
            IF (avg_db <= mindb) OR
               (avg_cf >= cfx)
            THEN BEGIN
                last_uv := vuv_pos;
                mVU := nVU DIV 2;
            END;

            IF lpc_pos+nLPC <= vuv_pos THEN BEGIN
                IF lpc_pos < last_uv THEN
                    nlp := last_uv - lpc_pos
                ELSE
                    nlp := nLPC;
                LoadBufferValuesDif(wrk_buf,sig,lpc_pos,nlp,scale,alpha);
                CovarianceLPC(sig,pcoef,nlp,ncoef,1000.0,ssqer);
                LoadBufferValuesDif(wrk_buf,sig,lpc_pos,nlp,1.0,0);
                dmp := 1.0;
                DC_scale := pcoef[0];
                FOR i := 1 TO ncoef DO BEGIN
                    dmp := dmp * bz;
                    pcoef[i] := pcoef[i]*dmp;
                    DC_scale := DC_scale + pcoef[i];
                END;
                IF DC_scale <> 0 THEN BEGIN
                    DC_scale := 1.0/DC_scale;
                    FOR i := 0 TO ncoef DO
                        pcoef[i] := DC_scale * pcoef[i];
                END;
                IF lpi_top < lpc_pos THEN
                    WHILE lpi_top < lpc_pos + nVU DIV 2 DO BEGIN
                        Inc(lpi_top);
                        lpi_buf[lpi_top] := 0;
                    END;
                FOR i := 0 TO lpi_top-lpc_pos DO BEGIN
                    c := Cos(Pi*i/Succ(lpi_top - lpc_pos));
                    lpi_buf[lpc_pos+i] := Round(
                        0.5*(1 + c)*lpi_buf[lpc_pos+i] +
                        0.5*(1 - c)*RevProduct(
                            sig[i],pcoef[0],Succ(ncoef))
                    );
                END;
                FOR i := Succ(lpi_top-lpc_pos) TO Pred(nlp) DO
                    lpi_buf[lpc_pos+i] :=
                    Round(RevProduct(sig[i],pcoef[0],Succ(ncoef)));
                lpi_top := Pred(lpc_pos+nlp);
                lpc_pos := lpc_pos + nlp DIV 2;
                IF last_uv < vuv_pos THEN BEGIN
                    mVU := nLPC DIV 2;
                END;
            END;

            itg_len := Pred(lpc_pos);
            IF itg_len > wrk_len THEN
                itg_len := wrk_len;
            WHILE lpi_top < itg_len DO BEGIN
                Inc(lpi_top);
                lpi_buf[lpi_top] := 0;
            END;
        END;

        IF nread = 0 THEN                  {On last time through}
            itg_len := lpi_top;

{Integration filtering, min finding, candidate pitch pulse evaluation}
        WHILE itg_top < itg_len DO BEGIN
            Inc(itg_top);
            IF lpi_buf[itg_top] <> 0 THEN
                itg_nzi := itg_top;
            IF itg_top > pos_ref+nLPC THEN BEGIN
                pos_ref := 0;
                 km_ref := 0;
            END;

{Apply integrators (matched-filter correlators)}
            x0 := lpi_buf[itg_top];
            IF pulse_pending THEN
                FOR k := 1 TO kmax DO BEGIN
                    y0[k] := b0[k]*(x0-x2) -a1[k]*y1[k] -a2[k]*y2[k];
                    y2[k] := y1[k];
                    y1[k] := y0[k];
                END
            ELSE
                FOR k := 1 TO kmax DO BEGIN
                    y0[k] := b0[k]*(x0-x2) -a1[k]*y1[k] -a2[k]*y2[k];
                    y2[k] := y1[k];
                    y1[k] := y0[k];
                    IF {adaptation restrictions}
                       ((km_ref = 0) OR
                        ((km_ref-2 <= k) AND (k <= km_ref+2)) OR
                        ((km_ref > 0) AND
                         (-a1[k]*b0[k]*data_scale*y0[k] < 2*val_opt))) AND
                       (y0[k] < ppm_val[1])
                    THEN BEGIN
                        ppm_val[1] := y0[k];
                        ppm_km [1] := k;
                        ppm_pos[1] := itg_top;
                    END;
                END;
            x2 := x1;
            x1 := x0;

{Optional integrator (correlator) output}
            IF out_mode = cor_out THEN BEGIN
                IF ppm_km[1] > 0 THEN
                    lpi_buf[itg_top] := Round(
                        -a1[ppm_km[1]]*b0[ppm_km[1]]*y0[ppm_km[1]])
                ELSE
                    lpi_buf[itg_top] := 0;
            END;

{Zero crossing following pulse detection - evaluate pulse candidates}
            IF (NOT pulse_pending) AND
               (ppm_km[1] > 0)     AND
               (y0[ppm_km[1]] > 0)
            THEN BEGIN
                itg_zcr := itg_top;
                pulse_pending := TRUE;
                lpisq := 0;
            END ELSE BEGIN
                x0 := data_scale*x0;
                lpisq := lpisq + x0*x0;
            END;
            IF pulse_pending       AND
               (ppm_km[1] > 0)     AND
               ((y0[ppm_km[1]] < 0.15*ppm_val[1]) OR
                (itg_top >= itg_nzi+nVU))
            THEN BEGIN
                pulse_pending := FALSE;
                ppm_val[1] := -a1[ppm_km[1]]*b0[ppm_km[1]]*
                    data_scale*ppm_val[1];
                p := pmax;
                WHILE (ppm_pos[p] = 0) OR
                      (ppm_pos[1] > ppm_pos[p] + nLPC)
                DO
                    Dec(p);
                 km_ref := ppm_km[1];
                pos_ref := ppm_pos[1];
                val_opt := 0;
                FOR ip := 1 TO p DO BEGIN
                    IF ppm_val[ip] < val_opt THEN
                        val_opt := ppm_val[ip];
                END;
                IF (itg_top >= itg_nzi+nVU) AND
                   (itg_zcr > itg_nzi)
                THEN
                    ppm_ok[1] := FALSE
                ELSE
                    ppm_ok[1] := TRUE;
                IF (itg_zcr >= itg_top) OR
                   (ppm_val[1]*ppm_val[1] < 1.4*lpisq/(itg_top-itg_zcr))
                THEN
                    ppm_ok[1] := FALSE;
                FOR ip := 2 TO p DO
                    IF ppm_val[ip] > 0.2*ppm_val[1] THEN
                        ppm_ok[ip] := FALSE;
                FOR ip := 1 TO Pred(p) DO
                    IF ppm_val[ip] > 0.2*ppm_val[p] THEN
                        ppm_ok[ip] := FALSE;
                FOR ip := 3 TO p DO BEGIN
                    val_pre := ppm_val[1];
                    val_pst := ppm_val[ip];
                    IF val_pre < val_pst THEN
                        val_pre := val_pst;
                    val_pre := 0.4*val_pre;
                    FOR im := 2 TO Pred(ip) DO
                        IF ppm_val[im] > val_pre THEN
                            ppm_ok[im] := FALSE;
                END;
                OutputPulse(pmax);
                FOR p := pmax DOWNTO 2 DO BEGIN
                    ppm_val[p] := ppm_val[Pred(p)];
                    ppm_pos[p] := ppm_pos[Pred(p)];
                    ppm_km [p] := ppm_km [Pred(p)];
                    ppm_ok [p] := ppm_ok [Pred(p)];
                END;
                ppm_val[1] := 0;
                ppm_pos[1] := 0;
                ppm_km [1] := 0;
                ppm_ok [1] := FALSE;
            END;
        END;
    UNTIL nread = 0;

{Clear out buffers}
    IF out_mode > 0 THEN BEGIN
        FOR i := Succ(lpi_top) TO wrk_len DO
            lpi_buf[i] := 0;
        IF wrk_len > 0 THEN
            StdBlockWrite(lpi_buf[1],wrk_len*Sizeof(INTEGER));
        StdSuspendWrite;
    END;
    FOR p := pmax DOWNTO 1 DO
        OutputPulse(p);
END;


BEGIN
    StdSetErrorExit;

    out_mode   := 0;
    mindb      := -40;
    cfx        := 1.0;
    fg         := 0.050;
    bg         := 0.0;
    bz         := 0.025;
    window_ms  := 20.0;
    fname      := '';
    extension  := '';

    FOR ipar := 1 TO ParamCount DO BEGIN
        pstr := ParamStr(ipar);
        IF pstr = '/lpi' THEN out_mode := lpi_out;
        IF pstr = '/cor' THEN out_mode := cor_out;
        UpdateSingleKeyword(fg,pstr,'/F:');
        UpdateSingleKeyword(bg,pstr,'/B:');
        UpdateSingleKeyword(bz,pstr,'/Z:');
        UpdateSingleKeyword(mindb,pstr,'/D:');
        UpdateStringKeyword(extension,pstr,'.');
        IF (fname = '')     AND
           (extension = '') AND
           (Copy(pstr,1,1) <> '/')
        THEN
            fname := pstr;
    END;
    IF bg <= 0 THEN
        bg := 2*fg;

    StdCmd('leader');
    IF extension <> '' THEN
        StdCmd(Concat('extension=',extension));
    StdCmd('input');
    IF (fname = '') AND (extension <> '') THEN
        fname := StdTxt('fname?');
    IF out_mode = lpi_out THEN BEGIN
        StdCmd('extension=lpi');
        StdCmd('output');
    END;
    IF out_mode = cor_out THEN BEGIN
        StdCmd('extension=cor');
        StdCmd('output');
    END;
    IF fname = '' THEN
        fname := StdTxt('ename?');
    sample_ms := StdDbl('sample_ms?');

    IF (sample_ms > 0) AND (fname <> '') THEN BEGIN
        start_ms   := StdDbl('start_ms?');
        scale      := StdDbl('scale?');
        data_scale := StdDbl('data_scale?');
        ncoef      := StdInt('ncoef?');
        alpha      := StdDbl('alpha?');
        bz         := Exp(-2*Pi*bz*sample_ms);

        Assign(pfile,fname);
        {$I-} Rewrite(pfile); {$I+}
        IF IOResult = 0 THEN BEGIN
            StdCmd('enable_break');
            StdCmd('');
            LPCInverseFilter;
            Close(pfile);
        END;
    END;
END.
