PROGRAM VoiceLP; {Copyright 1996, 1997 Paul Milenkovic ALL RIGHTS RESERVED}


USES StrBlock, StdIO, DSPTypes, DSPFast, LPC, FiltBuf;

VAR
    sample_ms, start_ms, stop_ms: DOUBLE;
    ncoef, i: INTEGER;
    scale, alpha, data_scale: DOUBLE;

    pstr, gname, oname, extension: ASCIIZ_string;
    gfile, ofile: TEXT;

    wrk_buf: FiltBuf.wrk_bufT;
    sig, sigp: DSPTypes.sig_buffer;


PROCEDURE LPCAnalysis;

    VAR
        wrk_len, wrk_dif, wrk_p1, wrk_pp, wrk_p2, nread, nbuf, i: INTEGER;

        tpm, tp, tpp, t0, dt1, dt2, rbuf: DOUBLE;
        w_max, amp, ampp, freq, freqp, bw, bwp: DOUBLE;
        ssqer, dmp: SINGLE;

        pcoef: predictor;

BEGIN
    wrk_len := 0;
    w_max := 20.0;
    t0 := start_ms;
    tpm := start_ms - w_max;
    IF NOT EOF(gfile) THEN
        Readln(gfile,tp,amp,freq,bw)
    ELSE
        Exit;
    IF NOT EOF(gfile) THEN
        Readln(gfile,tpp,ampp,freqp,bwp)
    ELSE
        tpp := stop_ms + w_max;

    REPEAT
        IF wrk_len > wrk_mrk THEN BEGIN
            wrk_dif := wrk_len - wrk_mrk;
            Move(wrk_buf[Succ(wrk_dif)],wrk_buf[1],wrk_mrk*Sizeof(INTEGER));
            Dec(wrk_len,wrk_dif);
            t0 := t0 + wrk_dif * sample_ms;
        END;
        nread := StdBlockRead(
            wrk_buf[Succ(wrk_len)],Sizeof(wrk_buf)-Sizeof(INTEGER)*wrk_len
        );
        wrk_len := wrk_len + nread DIV Sizeof(INTEGER);

        dt2 := tpp - tp;
        IF dt2 > w_max THEN dt2 := 0.5*w_max;
        WHILE (tp <= stop_ms) AND
              ((nread = 0) OR
               (tp + dt2 <= t0 + Pred(wrk_len)*sample_ms))
        DO BEGIN
            IF tp > t0 THEN BEGIN
                dt1 := tp - tpm;
                IF dt1 > w_max THEN dt1 := 0.5*w_max;
                IF (dt1 < w_max) AND (dt2 < w_max) THEN
                    rbuf := 0.5*(dt1 + dt2)/sample_ms
                ELSE
                IF dt1 < w_max THEN
                    rbuf := dt1/sample_ms
                ELSE
                    rbuf := dt2/sample_ms;
                nbuf := Round(rbuf);
                IF nbuf > mbuf THEN
                    nbuf := mbuf;
                wrk_p1 := Succ(Round((tp - 0.25*dt1 - t0)/sample_ms));
                wrk_pp := Succ(Round((tp - t0)/sample_ms));
                wrk_p2 := Succ(Round((tp + 0.75*dt2 - t0)/sample_ms));
                IF wrk_p1 < 2 THEN
                    wrk_p1 := 2;
                IF wrk_p2 > wrk_len THEN
                    wrk_p2 := wrk_len;
                IF wrk_p2 - wrk_p1 >= mbuf THEN
                    wrk_p2 := wrk_p1 + Pred(mbuf);
                IF (wrk_p1 <= wrk_p2) AND
                   (RMSdB(wrk_buf,wrk_p1,Succ(wrk_p2-wrk_p1),scale) > -90)
                THEN BEGIN {LPC}
                    LoadBufferValuesDif(
                        wrk_buf,sig,wrk_p1,Succ(wrk_p2-wrk_p1),scale,alpha
                    );
                    CovarianceLPC(
                        sig,pcoef,Succ(wrk_p2-wrk_p1),ncoef,1000.0,ssqer
                    );
                    Write(ofile,tp:0:3,#9,amp:0:6,#9,freq:0:3,#9,bw:0:3);
                    IF LPCtoLog2Area(pcoef,ncoef) THEN FOR i := 1 TO ncoef DO
                        Write(ofile,#9,pcoef[i]:0:6)
                    ELSE BEGIN
                        IF wrk_pp-nbuf < 2 THEN
                            nbuf := wrk_p1-2;
                        IF Pred(wrk_pp+nbuf) > wrk_len THEN
                            nbuf := Succ(wrk_len-wrk_pp);
                        LoadBufferValuesDif(
                            wrk_buf,sig,wrk_pp,nbuf,scale,alpha
                        );
                        LoadBufferValuesDif(
                            wrk_buf,sigp,wrk_pp-nbuf,nbuf,scale,alpha
                        );
                        FOR i := -ncoef TO Pred(nbuf) DO
                            sig[i] := (0.5 + 0.5*Cos(Pi*i/nbuf))*sig[i] +
                                      (0.5 - 0.5*Cos(Pi*i/nbuf))*sigp[i];
                        CovarianceLPC(
                            sig,pcoef,nbuf,ncoef,1000.0,ssqer
                        );
                        IF LPCtoLog2Area(pcoef,ncoef) THEN FOR i := 1 TO ncoef DO
                            Write(ofile,#9,pcoef[i]:0:6);
                     END;
                     Writeln(ofile);
                END {LPC};
            END;
            tpm := tp;
            tp := tpp;
            IF tp <= stop_ms THEN BEGIN
                amp := ampp;
                freq := freqp;
                bw := bwp;
                IF EOF(gfile) THEN
                    tpp := stop_ms + w_max
                ELSE
                    Readln(gfile,tpp,ampp,freqp,bwp);
                dt2 := tpp - tp;
                IF dt2 > w_max THEN dt2 := w_max;
            END;
        END;
    UNTIL (nread = 0) OR (tp > stop_ms);
END;


BEGIN
    StdSetErrorExit;

    gname      := '';
    extension  := '';

    FOR i := 1 TO ParamCount DO BEGIN
        pstr := ParamStr(i);
        UpdateStringKeyword(extension,pstr,'.');
        IF (gname = '')     AND
           (extension = '') AND
           (Copy(pstr,1,1) <> '/')
        THEN
            gname := pstr;
    END;

    StdCmd('leader');
    IF extension <> '' THEN
        StdCmd(Concat('extension=',extension));
    StdCmd('input');
    IF (gname = '') AND (extension <> '') THEN
        gname := StdTxt('aname?');
    StdCmd('extension=gpl');
    oname := StdTxt('fname?');
    IF gname = '' THEN
        gname := StdTxt('ename?');

    sample_ms := StdDbl('sample_ms?');
    IF (sample_ms > 0) AND (gname <> '') AND (oname <> '') THEN BEGIN
        start_ms   := StdDbl('start_ms?');
        stop_ms    := StdDbl('stop_ms?');
        scale      := StdDbl('scale?');
        data_scale := StdDbl('data_scale?');
        ncoef      := StdInt('ncoef?');
        alpha      := StdDbl('alpha?');

        Assign(gfile,gname);
        {$I-} Reset(gfile); {$I+}
        IF IOResult = 0 THEN BEGIN
            Assign(ofile,oname);
            {$I-} Rewrite(ofile); {$I+}
            IF IOResult = 0 THEN BEGIN
                StdCmd('enable_break');
                StdCmd('');
                LPCAnalysis;
                Close(ofile);
            END;
            Close(gfile);
        END;
    END;
END.
